package org.apache.bcel.verifier.structurals;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.apache.bcel.generic.ATHROW;
import org.apache.bcel.generic.BranchInstruction;
import org.apache.bcel.generic.GotoInstruction;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.JsrInstruction;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.RET;
import org.apache.bcel.generic.ReturnInstruction;
import org.apache.bcel.generic.Select;
import org.apache.bcel.verifier.exc.AssertionViolatedException;
import org.apache.bcel.verifier.exc.StructuralCodeConstraintException;

public class ControlFlowGraph {
   private final Subroutines subroutines;
   private final ExceptionHandlers exceptionhandlers;
   private final Map<InstructionHandle, InstructionContext> instructionContexts;

   public ControlFlowGraph(MethodGen methodGen) {
      this(methodGen, true);
   }

   public ControlFlowGraph(MethodGen methodGen, boolean enableJustIceCheck) {
      this.instructionContexts = new HashMap();
      this.subroutines = new Subroutines(methodGen, enableJustIceCheck);
      this.exceptionhandlers = new ExceptionHandlers(methodGen);
      InstructionHandle[] instructionhandles = methodGen.getInstructionList().getInstructionHandles();
      InstructionHandle[] var4 = instructionhandles;
      int var5 = instructionhandles.length;

      for(int var6 = 0; var6 < var5; ++var6) {
         InstructionHandle instructionhandle = var4[var6];
         this.instructionContexts.put(instructionhandle, new InstructionContextImpl(instructionhandle));
      }

   }

   public InstructionContext contextOf(InstructionHandle inst) {
      InstructionContext ic = (InstructionContext)this.instructionContexts.get(inst);
      if (ic == null) {
         throw new AssertionViolatedException("InstructionContext requested for an InstructionHandle that's not known!");
      } else {
         return ic;
      }
   }

   public InstructionContext[] contextsOf(InstructionHandle[] insts) {
      InstructionContext[] ret = new InstructionContext[insts.length];
      Arrays.setAll(ret, (i) -> {
         return this.contextOf(insts[i]);
      });
      return ret;
   }

   public InstructionContext[] getInstructionContexts() {
      InstructionContext[] ret = new InstructionContext[this.instructionContexts.size()];
      return (InstructionContext[])this.instructionContexts.values().toArray(ret);
   }

   public boolean isDead(InstructionHandle i) {
      return this.subroutines.subroutineOf(i) == null;
   }

   private class InstructionContextImpl implements InstructionContext {
      private int TAG;
      private final InstructionHandle instruction;
      private final Map<InstructionContext, Frame> inFrames;
      private final Map<InstructionContext, Frame> outFrames;
      private List<InstructionContext> executionPredecessors;

      public InstructionContextImpl(InstructionHandle inst) {
         if (inst == null) {
            throw new AssertionViolatedException("Cannot instantiate InstructionContextImpl from NULL.");
         } else {
            this.instruction = inst;
            this.inFrames = new HashMap();
            this.outFrames = new HashMap();
         }
      }

      private InstructionHandle[] _getSuccessors() {
         InstructionHandle[] single = new InstructionHandle[1];
         Instruction inst = this.getInstruction().getInstruction();
         InstructionHandle[] ret;
         if (inst instanceof RET) {
            Subroutine s = ControlFlowGraph.this.subroutines.subroutineOf(this.getInstruction());
            if (s == null) {
               throw new AssertionViolatedException("Asking for successors of a RET in dead code?!");
            } else {
               ret = s.getEnteringJsrInstructions();
               InstructionHandle[] retx = new InstructionHandle[ret.length];
               Arrays.setAll(retx, (i) -> {
                  return ret[i].getNext();
               });
               return retx;
            }
         } else if (!(inst instanceof ReturnInstruction) && !(inst instanceof ATHROW)) {
            if (inst instanceof JsrInstruction) {
               single[0] = ((JsrInstruction)inst).getTarget();
               return single;
            } else if (inst instanceof GotoInstruction) {
               single[0] = ((GotoInstruction)inst).getTarget();
               return single;
            } else if (inst instanceof BranchInstruction) {
               InstructionHandle[] pair;
               if (inst instanceof Select) {
                  pair = ((Select)inst).getTargets();
                  ret = new InstructionHandle[pair.length + 1];
                  ret[0] = ((Select)inst).getTarget();
                  System.arraycopy(pair, 0, ret, 1, pair.length);
                  return ret;
               } else {
                  pair = new InstructionHandle[]{this.getInstruction().getNext(), ((BranchInstruction)inst).getTarget()};
                  return pair;
               }
            } else {
               single[0] = this.getInstruction().getNext();
               return single;
            }
         } else {
            return InstructionHandle.EMPTY_ARRAY;
         }
      }

      public boolean execute(Frame inFrame, ArrayList<InstructionContext> execPreds, InstConstraintVisitor icv, ExecutionVisitor ev) {
         List<InstructionContext> clone = (List)execPreds.clone();
         this.executionPredecessors = clone;
         if ((this.lastExecutionJSR() != null || ControlFlowGraph.this.subroutines.subroutineOf(this.getInstruction()) == ControlFlowGraph.this.subroutines.getTopLevel()) && (this.lastExecutionJSR() == null || ControlFlowGraph.this.subroutines.subroutineOf(this.getInstruction()) != ControlFlowGraph.this.subroutines.getTopLevel())) {
            Frame inF = (Frame)this.inFrames.get(this.lastExecutionJSR());
            if (inF == null) {
               this.inFrames.put(this.lastExecutionJSR(), inFrame);
               inF = inFrame;
            } else if (inF.equals(inFrame) || !this.mergeInFrames(inFrame)) {
               return false;
            }

            Frame workingFrame = inF.getClone();

            try {
               icv.setFrame(workingFrame);
               this.getInstruction().accept(icv);
            } catch (StructuralCodeConstraintException var9) {
               StructuralCodeConstraintException ce = var9;
               ce.extendMessage("", "\nInstructionHandle: " + this.getInstruction() + "\n");
               ce.extendMessage("", "\nExecution Frame:\n" + workingFrame);
               this.extendMessageWithFlow(ce);
               throw ce;
            }

            ev.setFrame(workingFrame);
            this.getInstruction().accept(ev);
            this.outFrames.put(this.lastExecutionJSR(), workingFrame);
            return true;
         } else {
            throw new AssertionViolatedException("Huh?! Am I '" + this + "' part of a subroutine or not?");
         }
      }

      private void extendMessageWithFlow(StructuralCodeConstraintException e) {
         String s = "Execution flow:\n";
         e.extendMessage("", "Execution flow:\n" + this.getExecutionChain());
      }

      public ExceptionHandler[] getExceptionHandlers() {
         return ControlFlowGraph.this.exceptionhandlers.getExceptionHandlers(this.getInstruction());
      }

      private String getExecutionChain() {
         StringBuilder s = new StringBuilder(this.toString());

         for(int i = this.executionPredecessors.size() - 1; i >= 0; --i) {
            s.insert(0, this.executionPredecessors.get(i) + "\n");
         }

         return s.toString();
      }

      public Frame getInFrame() {
         InstructionContext jsr = this.lastExecutionJSR();
         Frame org = (Frame)this.inFrames.get(jsr);
         if (org == null) {
            throw new AssertionViolatedException("inFrame not set! This:\n" + this + "\nInFrames: '" + this.inFrames + "'.");
         } else {
            return org.getClone();
         }
      }

      public InstructionHandle getInstruction() {
         return this.instruction;
      }

      public Frame getOutFrame(ArrayList<InstructionContext> execChain) {
         this.executionPredecessors = execChain;
         InstructionContext jsr = this.lastExecutionJSR();
         Frame org = (Frame)this.outFrames.get(jsr);
         if (org == null) {
            throw new AssertionViolatedException("outFrame not set! This:\n" + this + "\nExecutionChain: " + this.getExecutionChain() + "\nOutFrames: '" + this.outFrames + "'.");
         } else {
            return org.getClone();
         }
      }

      public InstructionContext[] getSuccessors() {
         return ControlFlowGraph.this.contextsOf(this._getSuccessors());
      }

      public int getTag() {
         return this.TAG;
      }

      private InstructionContextImpl lastExecutionJSR() {
         int size = this.executionPredecessors.size();
         int retcount = 0;

         for(int i = size - 1; i >= 0; --i) {
            InstructionContextImpl current = (InstructionContextImpl)this.executionPredecessors.get(i);
            Instruction currentlast = current.getInstruction().getInstruction();
            if (currentlast instanceof RET) {
               ++retcount;
            }

            if (currentlast instanceof JsrInstruction) {
               --retcount;
               if (retcount == -1) {
                  return current;
               }
            }
         }

         return null;
      }

      private boolean mergeInFrames(Frame inFrame) {
         Frame inF = (Frame)this.inFrames.get(this.lastExecutionJSR());
         OperandStack oldstack = inF.getStack().getClone();
         LocalVariables oldlocals = inF.getLocals().getClone();

         try {
            inF.getStack().merge(inFrame.getStack());
            inF.getLocals().merge(inFrame.getLocals());
         } catch (StructuralCodeConstraintException var6) {
            StructuralCodeConstraintException sce = var6;
            this.extendMessageWithFlow(sce);
            throw sce;
         }

         return !oldstack.equals(inF.getStack()) || !oldlocals.equals(inF.getLocals());
      }

      public void setTag(int tag) {
         this.TAG = tag;
      }

      public String toString() {
         return this.getInstruction().toString(false) + "\t[InstructionContext]";
      }
   }
}
